//
//  main.cpp
//  SynthEco
//


/*dependencies
 cpny https://github.com/rogersce/cnpy
 libsndfile
*/

#include <iostream>
//#include <armadillo>  //need this later?
#include <algorithm>
#include <iterator>
#include <array>
#include <list>
#include <cmath>
#include "maximilian.h"
#include "maxiFFT.h"
#include "RandLib.hpp"
#include "SoundUtil.hpp"
#include"cnpy.h"
#include <thread>
using namespace std;
//using namespace arma;

template < typename T > void shuffle( std::list<T>& lst )
{
    // create a vector of (wrapped) references to elements in the list
    // http://en.cppreference.com/w/cpp/utility/functional/reference_wrapper
    std::vector< std::reference_wrapper< const T > > vec( lst.begin(), lst.end() ) ;
    
    // shuffle (the references in) the vector
    std::shuffle( vec.begin(), vec.end(), std::mt19937{ std::random_device{}() } ) ;
    
    // copy the shuffled sequence into a new list
    std::list<T> shuffled_list {  vec.begin(), vec.end() } ;
    
    // swap the old list with the shuffled list
    lst.swap(shuffled_list) ;
}


#define DEF_MAXBARMASKLENGTH 8

struct synthEcoLib {
    
    enum geneDescriptors : int {
        PITCH=0,
        TIMESLOT,
        PERIOD,
        LISTENING_BANDWIDTH,
        BARMASK,
        BARMASK_LENGTH,
        PAN, //this is decorative and has no effect on the model
        LAST_GENE_MARKER
    };

    
    static maxiFFT fft;
    
    struct stats {
        vector<int> populationSize;
        vector<int> periods;
        vector<int> masks;
        vector<int> energies;
        vector<int> births;
        vector<int> deaths;
        vector<float> averageAge;
    };
    
    struct agent {
        int energy;
        vector<double> g;
        int id=-1;
        maxiOsc osc;
        vector<double> agentSound;
        vector<float> agentMags;
        int periodLength;
        unsigned int barMask;
        int barMaskLength;
        unsigned int age=0;
    };
    typedef list<agent> agentList;
    typedef std::reference_wrapper<agent> agentRef;
    typedef list<agentRef > agentRefList;
    typedef vector<agentRefList > agentRefListVect;
    

    
    struct worldState {
        stats runstats;
        agentList pop;
        int initPop=500;
        int maxPop = 1000;
        int maxPeriodLength = 100;
        int pitchRange = 96;
        double initEnergyMean = 220;
        double initEnergyStddev = 50;
        int energyTaxForNonParents = 5;
        int energyBoostForCommunication = 17;
        int energyTaxForReproduction = 50;
        int energyThresholdForReproduction = 300;
        int energyTaxForOvercrowding = 250;
        int energyTaxForVocalising = 5;
        double mutateProb = 0.1;
        double mutateScale = 0.15;
        double pitchMutateProb = 0.05;
        double pitchMutateScale = 0.15;
        unsigned int age = 0;
        unsigned int maxAge = 200000;
        double reproductionProb = 0.5;
        vector<double> worldSound, worldSound2ch;
        vector<float> worldMags;
        unsigned int timeStepLength = 512; // in samples
        int maxListeningBandwidth = 50;
        int minListeningBandwidth = 20;
        int maxBarMaskLength = DEF_MAXBARMASKLENGTH;
        
        maxiFilter filter;
        
        bool minimalMode = 1;
        bool nullModelMode = 1;
        
    };
    
    static agent initAgent (double energyMean, double energyStd, unsigned int geneSize) {
        static int id=0;
        agent a;
        a.energy = max(1,(int)round(RandLib::inst().randn(energyMean, energyStd)));
        a.g = RandLib::inst().uvec(geneSize);
        a.id = id++;
        return a;
    };
    

    static worldState createPop (worldState&& world, int geneSize) {
        world.pop.resize(world.initPop); //built in stl vector method
        for(agent& v : world.pop) {
            v = initAgent(world.initEnergyMean, world.initEnergyStddev, geneSize);
        }
        return std::move(world);
    };
    
    static double mutate (const double value, double mutateProbability, double mutateScale) {
        double mutatedValue = value;
        if (RandLib::inst().randu<double>(0,1) < mutateProbability) {
            mutatedValue = mutatedValue + RandLib::inst().randn<double>(0,mutateScale);
            double intpart;
            if (mutatedValue > 1.0) {
                mutatedValue = modf(mutatedValue,&intpart);
            }else if (mutatedValue < 0.0) {
                mutatedValue = 1.0 + modf(mutatedValue,&intpart);
            }
        }
        return mutatedValue;
    };
   
    static agent reproduceAsexual (const agent &a1, double initEnergyMean, double initEnergyStd, const double mutateProb, const double mutateScale, const double pitchMutateProb, const double pitchMutateScale) {
        agent newAgent;
        newAgent = initAgent(initEnergyMean, initEnergyStd, (int)a1.g.size());
        newAgent.g.at(geneDescriptors::PITCH) = synthEcoLib::mutate(a1.g.at(geneDescriptors::PITCH), pitchMutateProb, pitchMutateScale);
        newAgent.g.at(geneDescriptors::TIMESLOT) = synthEcoLib::mutate(a1.g.at(geneDescriptors::TIMESLOT), mutateProb, mutateScale );
        newAgent.g.at(geneDescriptors::PERIOD) = synthEcoLib::mutate(a1.g.at(geneDescriptors::PERIOD), mutateProb, mutateScale );
        newAgent.g.at(geneDescriptors::LISTENING_BANDWIDTH) = synthEcoLib::mutate(a1.g.at(geneDescriptors::LISTENING_BANDWIDTH), mutateProb, mutateScale );
        newAgent.g.at(geneDescriptors::BARMASK) = synthEcoLib::mutate(a1.g.at(geneDescriptors::BARMASK), mutateProb, mutateScale );
        newAgent.g.at(geneDescriptors::BARMASK_LENGTH) = synthEcoLib::mutate(a1.g.at(geneDescriptors::BARMASK_LENGTH), mutateProb, mutateScale );
        return newAgent;
    };

    static int geneToPitch (int pitchRange, double v) {
        return int(v*pitchRange) + 1;
    };

    static int geneToTimeslot (int period, double v) {
        return (int)(v*period);
    };

    static int geneToPeriod (int nTimeslots, double v) {
        return (int)(v*(nTimeslots-1)) + 1;
    };
    
    static int geneToListeningBandwidth(int minBandwidth, int maxBandwidth, double v) {
        return (int)(v*(maxBandwidth - minBandwidth)) + minBandwidth;
    }

    static unsigned int geneToBarMask(int maxBarMaskLength, double v) {
        return (unsigned int)((v * pow(2, maxBarMaskLength)) + 1);
    }

    static int geneToBarMaskLength(int maxBarMaskLength, double v) {
        return (int)((v * maxBarMaskLength) + 1);
    }

    static void logPopultation(const worldState& world) {
        for(auto &v : world.pop) {
            cout << "Agent: \t" << v.id << "\t" << v.energy << "\t" << geneToPitch(world.pitchRange, v.g[0]) << "\t" << geneToTimeslot(world.maxPeriodLength, v.g[1]) << endl;
        }
    };
    
    static array<double,3> worldSoundSynthesis(const agentRefListVect & pitchCells, const double phase) {
        double w = 0.0, wL = 0.0, wR = 0.0;
        for(auto &v: pitchCells) {
            for(auto &agent : v) {
                double agentw = agent.get().osc.sinewave(200. + ((geneToPitch((int)pitchCells.size(), agent.get().g[0]) - 1) * 100));
                double amp = sin(phase * PI);
                agent.get().agentSound.push_back(agentw);
                agentw *= amp; //envelope after pushing?
                w+= agentw;
                double pan = agent.get().g[geneDescriptors::PAN];
                wL = w * pan;
                wR = w * (1.0 - pan);
            }
        }
        w *= 0.1;
        array<double, 3> output = {w, wL * 0.1, wR * 0.1};
        return output;
    }
    
    static array<double, 3> worldSoundEvents(worldState & world, array<double, 3> &w) {
        double progress = world.age / (double) world.maxAge;
        double noise=0.0;
        if (progress > 0.2 && progress < 0.25) {
            noise = world.filter.lores(RandLib::inst().randu(-0.5,0.5), 2000, 1);
        }
        w[0] += noise;
        w[1] += (noise/2.0);
        w[2] += (noise/2.0);
        return w;
    }

    static worldState timeStep(worldState && world) {
        
        cout <<"Update, age: " << world.age << ", pop size: " << world.pop.size() << endl;

        agentRefListVect niches(world.pitchRange);
        for(agent& v : world.pop) {
            if (world.nullModelMode) {
                v.g[geneDescriptors::PERIOD] = RandLib::inst().randu<double>(0,1);
                v.g[geneDescriptors::TIMESLOT] = RandLib::inst().randu<double>(0,1);
                v.g[geneDescriptors::BARMASK] = RandLib::inst().randu<double>(0,1);
                v.g[geneDescriptors::BARMASK_LENGTH] = RandLib::inst().randu<double>(0,1);
            }
            v.periodLength = world.minimalMode ? world.maxPeriodLength : geneToPeriod(world.maxPeriodLength, v.g[geneDescriptors::PERIOD]);
            int ts = geneToTimeslot(world.minimalMode ? world.maxPeriodLength : v.periodLength, v.g[geneDescriptors::TIMESLOT]);
            unsigned int barMask = geneToBarMask(world.maxBarMaskLength, v.g[geneDescriptors::BARMASK]);
            int barMaskLength = geneToBarMaskLength(world.maxBarMaskLength, v.g[geneDescriptors::BARMASK_LENGTH]);
            int bar = (int) world.age / v.periodLength;
            int barPhase = bar % barMaskLength;
            auto barMaskBits = bitset<DEF_MAXBARMASKLENGTH>(barMask);
            if ((barMaskBits.test(barPhase) || world.minimalMode) &&  world.age % v.periodLength == ts) {
                if (world.nullModelMode) {
                    v.g[geneDescriptors::PITCH] = RandLib::inst().randu<double>(0,1);
                }
                int pitch = geneToPitch(world.pitchRange, v.g[geneDescriptors::PITCH]);
                niches[pitch-1].push_back(v);
                float tax =  world.energyTaxForVocalising;
                v.energy -= tax;
            }
        }
        
        //sound
        //clear out agent sound buffers
        for(auto &v: world.pop) {
            v.agentSound.clear();
        };
        
        //calc agents and world sounds
        for(int sample=0; sample < world.timeStepLength; sample++) {
            array<double, 3> output = worldSoundSynthesis(niches, sample / (double) world.timeStepLength);
            output = worldSoundEvents(world, output) ;
            world.worldSound.push_back(output[0]);
            world.worldSound2ch.push_back(output[1]);
            world.worldSound2ch.push_back(output[2]);
        }
        
        //listening agents
        //first, get fft of world sound
        for(unsigned long i=world.worldSound.size() - world.timeStepLength; i < world.worldSound.size(); i++) {
            if (fft.process(world.worldSound[i])) {
                world.worldMags = vector<float>(&fft.magnitudes[0], &fft.magnitudes[0] + fft.bins);
            }
        }
        
        //next, compare the agents
        //for each agent
        unsigned int birthCount=0;
        unsigned int deathCount=0;
        for(auto &v: niches) {
            for(auto &anAgent : v) {
                //fft of its sound
                for(double w: anAgent.get().agentSound) {
                    if(fft.process(w)) {
                        anAgent.get().agentMags = vector<float>(&fft.magnitudes[0], &fft.magnitudes[0] + fft.bins);
                        
                        //subtract agentMags from worldMags
                        vector<float> fftDiff(fft.bins);
                        for(int k=0; k < fft.bins; k++) {
                            fftDiff[k] = world.worldMags[k] - anAgent.get().agentMags[k];
                        }
                        //what's the peak frequency in agentMags?
                        auto agentFFTPeakElement = std::max_element(anAgent.get().agentMags.begin(), anAgent.get().agentMags.end());
                        auto peakElementIndex = std::distance(anAgent.get().agentMags.begin(), agentFFTPeakElement);
                        
                        //how loud is this freqency area?
                        float areaAmp = 0;
                        int agentListeningBandwidth = world.minimalMode ? world.minListeningBandwidth : geneToListeningBandwidth(world.minListeningBandwidth, world.maxListeningBandwidth, anAgent.get().g[geneDescriptors::LISTENING_BANDWIDTH]);
                        for(long k=max(0l, peakElementIndex-agentListeningBandwidth);
                                      k < min(peakElementIndex + agentListeningBandwidth, (long) fft.bins);
                                      k++) {
                            areaAmp += fftDiff[k];
                        }
                        areaAmp /= (agentListeningBandwidth * 2.0);
//                            cout << areaAmp << endl;
                        double reproductionThreshold = world.minimalMode ? -6.1 : (-5.87 +(anAgent.get().g[geneDescriptors::LISTENING_BANDWIDTH] * 0.51));
                        if (areaAmp < reproductionThreshold) {
                            
                            anAgent.get().energy += world.energyBoostForCommunication;
                            //reproduce
                            if (anAgent.get().energy > world.energyThresholdForReproduction) {
                                if(world.pop.size() < world.maxPop) {
                                    if (RandLib::inst().randu<double>(0,1) < world.reproductionProb) {
                                        agent spawn;
                                        spawn = synthEcoLib::reproduceAsexual(anAgent.get(), world.initEnergyMean, world.initEnergyStddev, world.mutateProb, world.mutateScale, world.pitchMutateProb, world.pitchMutateScale);
                                        world.pop.push_back(spawn);
                                        anAgent.get().energy -= world.energyTaxForReproduction;
                                        birthCount++;
                                    }
                                }
                            }
                        }else if (areaAmp > -5.13) {
                            float tax =  world.energyTaxForOvercrowding;
                            anAgent.get().energy -= tax;
                        }
                        else{
                            float tax =  world.energyTaxForNonParents;
                            anAgent.get().energy -= tax;
                        }
                    }
                }
            }
        }

        cout << "Pop size after reproduction: " << world.pop.size() << endl;
        deathCount = (unsigned int) world.pop.size();
        //culling
        auto newEnd = remove_if(world.pop.begin(), world.pop.end(), [](const agent& a) {return a.energy <= 0;});
        world.pop.erase(newEnd, world.pop.end());
        
        cout << "Pop size after culling: " << world.pop.size() << endl;
        deathCount -= (unsigned int) world.pop.size();
        
        float avgAge = 0.0;
        for(agent& v : world.pop) {
            avgAge += v.age;
            v.age++;
            world.runstats.periods.push_back(v.periodLength);
        }
        for(int i=0; i < world.maxPop - world.pop.size(); i++) {
            world.runstats.periods.push_back(0);
        }
        avgAge /= world.pop.size();
        //stats
        world.runstats.populationSize.push_back((int) world.pop.size());
        world.runstats.births.push_back(birthCount);
        world.runstats.deaths.push_back(deathCount);
        world.runstats.averageAge.push_back(avgAge);
        
        world.age++;
        return std::move(world);
    };
    
    static worldState simulate(worldState&& world) {
        for(int i=0; i < world.maxAge && world.pop.size() > 0; i++) {
            world = timeStep(std::move(world));
        }
        return std::move(world);
    }
    
    static void saveStats(const worldState &world) {
        cnpy::npy_save("/tmp/popsize.npy", &world.runstats.populationSize[0], {world.runstats.populationSize.size()},"w");
        cnpy::npy_save("/tmp/births.npy", &world.runstats.births[0], {world.runstats.births.size()},"w");
        cnpy::npy_save("/tmp/deaths.npy", &world.runstats.deaths[0], {world.runstats.deaths.size()},"w");
        cnpy::npy_save("/tmp/averageAge.npy", &world.runstats.averageAge[0], {world.runstats.averageAge.size()},"w");
        cnpy::npy_save("/tmp/periods.npy", &world.runstats.periods[0], {(unsigned long)world.maxAge, (unsigned long)world.maxPop},"w");
    }
    
};

maxiFFT synthEcoLib::fft;


int main(int argc, const char * argv[]) {
    
    
    synthEcoLib::worldState world;
    synthEcoLib::fft.setup(512,512,512);
    world.worldSound.reserve(world.timeStepLength * world.maxAge);

    world = synthEcoLib::createPop(std::move(world), synthEcoLib::geneDescriptors::LAST_GENE_MARKER);
    cout << "Init pop size: " << world.pop.size()<< endl;;

    world = synthEcoLib::simulate(std::move(world));
    SoundUtil::writeWAV(world.worldSound, 1, "test.wav");
    SoundUtil::writeWAV(world.worldSound2ch, 2, "testSt.wav");
    synthEcoLib::saveStats(world);

    
    return 0;
}
